"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getModel = getModel;
exports.initDB = initDB;
const lodash_1 = __importDefault(require("lodash"));
const objection_1 = require("objection");
const constants_1 = require("./constants");
const dispatcher_1 = require("./dispatcher");
function getModel() {
    let _db;
    class CoreBaseModel extends objection_1.Model {
        static set db(value) {
            if (!_db) {
                _db = value;
            }
        }
        static get primary() {
            return _db.primary;
        }
        get primary() {
            return _db.primary;
        }
        static get replica() {
            return _db.replica;
        }
        get replica() {
            return _db.replica;
        }
    }
    return CoreBaseModel;
}
function initDB({ connectionString, dispatcherOptions, knexOptions = {}, logger = constants_1.defaultExLogger, }) {
    if (!connectionString) {
        throw new Error('Empty connection string');
    }
    const mergedKnexOptions = lodash_1.default.merge({}, constants_1.defaultKnexOptions, knexOptions);
    const mergedDispatcherOptions = lodash_1.default.merge({}, constants_1.defaultDispatcherOptions, dispatcherOptions);
    const db = new dispatcher_1.PGDispatcher({
        connections: connectionString.split(','),
        options: mergedDispatcherOptions,
        knexOptions: mergedKnexOptions,
        logger,
    });
    const terminate = () => {
        const beforeTerminate = mergedDispatcherOptions.beforeTerminate;
        beforeTerminate()
            .catch(() => { })
            .then(() => db.terminate())
            .then(() => process.exit(0))
            .catch((error) => {
            // eslint-disable-next-line no-console
            console.error(error);
            process.exit(-1);
        });
    };
    process.on('SIGINT', terminate);
    process.on('SIGUSR1', terminate);
    process.on('SIGUSR2', terminate);
    const CoreBaseModel = getModel();
    CoreBaseModel.db = db;
    const helpers = {
        clearDatabase: async function () {
            await db.ready();
            await db.primary.migrate.rollback({}, true);
        },
        rollbackDatabase: async function (args) {
            await db.ready();
            const onlyOne = args ? args.onlyOne : false;
            if (onlyOne) {
                await db.primary.migrate.down();
            }
            else {
                await db.primary.migrate.rollback({});
            }
        },
        migrateDatabase: async function (args) {
            await db.ready();
            const onlyOne = args ? args.onlyOne : false;
            if (onlyOne) {
                await db.primary.migrate.up();
            }
            else {
                await db.primary.migrate.latest();
            }
        },
        prepareDatabase: async function () {
            await db.primary.migrate.latest();
            await db.primary.seed.run();
        },
    };
    logger.info('Core-db is up and running!');
    return { db, CoreBaseModel, helpers };
}
